    module SPTpol_2012_TEEE
    use cmbtypes
    use likelihood
    use MatrixUtils
    use settings

    implicit none


    ! The bandpowers, in order:
    !                 150x150TE, 150x150EE
    ! cov == The cholesky factored bandpower covariance 
    integer :: nfreq, nband, nbin, nall, bands_per_freq
    double precision, dimension(:,:), allocatable :: spec, cov, cov_sims, new_cov_data, cov_data, cov_scale
    double precision, dimension(:,:), allocatable :: windows, cl_derivative, raw_spectra, ells
    integer :: spt_windows_lmin, spt_windows_lmax
    double precision, dimension(:), allocatable :: cl_to_dl_conversion, dl_th, cal_vector
    
    logical :: SuccessfulSPTInitialization
    logical :: use_BB, do_wimp
    
    type, extends(CosmologyLikelihood) :: SPTpolLikelihood
     contains
       procedure :: LogLike => SPTpolLnLike
    end type SPTpolLikelihood

    contains

    subroutine SPTpolLikelihood_Add(LikeList, Ini)
      use IniFile
      use settings
      class(LikelihoodList) :: LikeList
      Type(TIniFile) :: ini
      Type(SPTpolLikelihood), pointer :: like

      character (LEN=Ini_max_string_len) :: desc_file
      character (LEN=Ini_max_string_len) :: bp_file
      character (LEN=Ini_max_string_len) :: cov_data_file, cov_sims_file
      character (LEN=Ini_max_string_len) :: win_folder

      if (.not. Ini_Read_Logical_File(Ini, 'use_SPTpol', .false.)) then
         return
      end if

      allocate(like)
      Like%speed = 10
      Like%LikelihoodType = 'SPTpol'
      Like%name = 'SPTpol_2012_TEEE'
      Like%needs_background_functions = .true.

      print *, 'Likelihood Name: ', Like%name

      call Like%loadParamNames(trim(DataDir)//'SPTpol.paramnames')
      call LikeList%Add(like)

      !I could simplify this by moving all the relevant info from
      !desc_file into the ini file.
      use_BB = Ini_Read_logical_File(Ini, 'use_BB', .false.)
      desc_file = Ini_Read_String_Default_File(Ini, 'sptpol_desc_file', &
           trim(DataDir)//'sptpol_2012.desc_file')

      !Read in the bandpower and cov files, and determine where the windows are.
      bp_file = Ini_Read_String_Default_File(Ini, 'sptpol_bp_file','')
      cov_data_file = Ini_Read_String_Default_File(Ini, 'sptpol_cov_data_file','')
      cov_sims_file = Ini_Read_String_Default_File(Ini, 'sptpol_cov_sims_file','')
      win_folder = Ini_Read_String_Default_File(Ini, 'sptpol_window_dir','')
      if (bp_file=='' .or. desc_file=='' .or. win_folder=='' .or. cov_data_file=='' .or. cov_sims_file=='') then
         print*,'Missing required sptpol key: received: ',bp_file,desc_file,win_folder,cov_data_file,cov_sims_file
         stop
      endif

      print *, 'Use BB?: ', use_BB

      call InitSPTpolData(desc_file, bp_file, cov_data_file, cov_sims_file, win_folder)
    end subroutine SPTpolLikelihood_Add

    subroutine InitSPTpolData(desc_file, bp_file, cov_data_file, cov_sims_file, win_folder)
      use IniFile
      character(LEN=Ini_max_string_len) :: desc_file, bp_file, cov_data_file, cov_sims_file
      character(LEN=Ini_max_string_len) :: win_folder
      integer i,j,k,dum,lwin
      real rtmp

      !Obtain necessary info from the desc_file pertaining
      !to which freqs we're using, ell range, and number of bins per spectrum.
      call OpenTxtFile(desc_file, tmp_file_unit)
      print *, 'Reading desc file...',trim(desc_file)

      read (tmp_file_unit,*) nbin,nfreq !Number of bandpowers per spectrum, Number of freqs
      read (tmp_file_unit,*) spt_windows_lmin, spt_windows_lmax !Min and Max ell in window files.
      close(tmp_file_unit)

      print *, 'nbin: ', nbin
      print *, 'nfreq: ', nfreq
      print *, 'spt_windows_lmin: ', spt_windows_lmin
      print *, 'spt_windows_lmax: ', spt_windows_lmax
      print *, 'win_folder: ', trim(win_folder)

      !If use_BB = False, we just have EE and TE, otherwise add BB for each freq.
      if (use_BB == .true.) then
         bands_per_freq = 4
      else
         bands_per_freq = 3! Should be two for SPTpol (3 with TT).
      end if

      !nband = bands_per_freq*(nfreq)*(nfreq+1)/2 !With cross-freq spectra.
      nband = bands_per_freq*(nfreq)
      nall = nbin*(nfreq * (bands_per_freq - 1)) !Cov shouldn't know about TT.

      print *, 'nband: ', nband
      print *, 'nall: ', nall
      print *, 'bands_per_freq: ', bands_per_freq

      if (lmax .lt. spt_windows_lmax) then
         spt_windows_lmax=lmax
      end if

      if (spt_windows_lmin < 2 .or. spt_windows_lmin >= spt_windows_lmax) then
         call mpistop('Invalid lranges for sptpol')
      end if
      
      if (spt_windows_lmin .lt. 2) then
         print *, "Unexpected lmin for windows!"
         stop
      end if

      allocate(cl_to_dl_conversion(spt_windows_lmin-1:spt_windows_lmax+1), &
           dl_th(spt_windows_lmin:spt_windows_lmax), &
           cal_vector(1:nall))

      allocate(ells(spt_windows_lmin-1:spt_windows_lmax+1, 1:3))

      allocate(windows(spt_windows_lmin:spt_windows_lmax,1:nall), &
           spec(1:nbin,nband))

      allocate(new_cov_data(1:nall,1:nall), cov_data(1:nall,1:nall),cov_sims(1:nall,1:nall), &
           cov(1:nall,1:nall),cov_scale(1:nall,1:nall))

      allocate(cl_derivative(spt_windows_lmin:spt_windows_lmax,1:3))
      allocate(raw_spectra(spt_windows_lmin-1:spt_windows_lmax+1,1:3))
      
      !Define an array with the l*(l+1)/2pi factor to convert to Dl from Cl.
      do j=spt_windows_lmin-1,spt_windows_lmax+1
         ells(j,1:3) = j
         cl_to_dl_conversion(j) = (1.0 + j)*(j / TWOPI)
      end do

      ! Read in bandpowers
      !Should be TE, EE, TT, in that order from SPTpol analysis.
      print *, 'Reading in bandpowers...'
      call OpenTxtFile(bp_file, tmp_file_unit)
      do i=1,nband
         do j=1,nbin
            read (tmp_file_unit,*) dum,spec(j,i)
         end do
      end do
      close(tmp_file_unit)
    
      ! Read in covariance
      !Should be TE, EE
      print *, 'Reading in cov_data matrix...'
      call OpenTxtFile(cov_data_file, tmp_file_unit)
      do i=1,nall
         do j=1,nall
            read (tmp_file_unit,*) cov_data(i,j)
         end do
      end do
      close (tmp_file_unit)

      print *, 'Reading in cov_sims matrix...'
      call OpenTxtFile(cov_sims_file, tmp_file_unit)
      do i=1,nall
         do j=1,nall
            read (tmp_file_unit,*) cov_sims(i,j)
         end do
      end do
      close (tmp_file_unit)
      
      ! Read in windows
      !Should be TE, EE
      print *, 'Reading in windows...'
      do i=1,nall
         call OpenTxtFile(trim(win_folder)//trim(numcat('/window_',i)),tmp_file_unit)
         do j=spt_windows_lmin,spt_windows_lmax
            read (tmp_file_unit,*) dum, windows(j,i)
         end do
         close(tmp_file_unit)
      end do

      SuccessfulSPTInitialization = .true.
    end subroutine InitSPTpolData

    function SPTpolLnLike(like, CMB, Theory, DataParams)
      class(SPTpolLikelihood) :: like
      class(CMBParams) :: CMB
      class(TheoryPredictions) :: Theory !cl(lmax, num_tot_cls) TT, TE, EE, BB,(+phi) in that order from Camb.
      real(mcp) :: SPTpolLnLike
      real(mcp) :: TcalLnLike
      real(mcp) :: kappaLnLike
      real(mcp) :: dum
      double precision :: DataParams(:) !Foreground parameters (plus wimp annihilation vector amplitude)
      double precision, parameter :: d3000 = 3000*3001/(2*pi)
      double precision, parameter :: sigmaTcal = 0.013 !Uncertainty from cross-map calibration.
      double precision, parameter :: meanTcal = 1.0 !Maps are pre-calibrated so we expect amplitude 1.0.
      double precision, parameter :: sigmakappa = 0.00245 !Uncertainty on kappa.
      double precision, parameter :: meankappa = 0.0 !Mean lensing convergence.
      double precision, dimension(1:nall) :: deltacb
      double precision, dimension(1:nbin) :: tmpcb
      double precision, dimension(1) :: junk
      integer :: i,j,k, l
      real(mcp) :: norm
    
      double precision, dimension(2:lmax) :: cl_fgs

      !First calculate Cl derivatives for this position in parameter space.
      !raw_spectra = l^3 C_l
      !CAMB theory curves are in Cl as read in by CosmoMC, not Dl!!! So we don't need to do the dl_to_cl_conversion.
      raw_spectra(spt_windows_lmin-1:spt_windows_lmax+1,1:3) = ells(spt_windows_lmin-1:spt_windows_lmax+1,1:3)**3 * Theory%cl(spt_windows_lmin-1:spt_windows_lmax+1, 1:3)

      !The derivative at point n is roughly (raw_spectra[n+1] - raw_spectra[n-1])/2.  Then we devide by l^2 for the proper scaling for the
      !kappa parameter as described in Manzotti, et al. 2014, equation (32).
      cl_derivative(spt_windows_lmin:spt_windows_lmax,1:3) = (raw_spectra(spt_windows_lmin+1:spt_windows_lmax+1,1:3) - raw_spectra(spt_windows_lmin-1:spt_windows_lmax-1,1:3))/2./ells(spt_windows_lmin:spt_windows_lmax,1:3)**2

      !DataParams for SPTpol likelihood are: [MapTcal, MapPcal, Czero_EE, CzeroTE, TTepsilon, kappa]
      i=0
      do j=1,(nfreq)
         do k=1,bands_per_freq - 1 !Don't care about TT. Just use it for leakage, which is turned off by default.
            i=i+1
          
            tmpcb(:)=0
          
            !First get model foreground spectrum (in Cl).
            !Note all the foregrounds are recorded in Dl at l=3000, so we 
            !divide by d3000 to get to a normalized Cl spectrum.
            if (k==1) then
               !k==1 is SPTpol TE spectrum.  Include TE Poisson foreground.
               cl_fgs(spt_windows_lmin:spt_windows_lmax) = DataParams(4)/d3000

               !Now subtract the kappa parameter.
               cl_fgs(spt_windows_lmin:spt_windows_lmax) = cl_fgs(spt_windows_lmin:spt_windows_lmax) - DataParams(6)*cl_derivative(spt_windows_lmin:spt_windows_lmax, 2)
            end if 

            if (k==2) then
               !k==2 is SPTpol EE spectrum. Include EE Poisson foreground.
               cl_fgs(spt_windows_lmin:spt_windows_lmax) = DataParams(3)/d3000 

               !Now subtract the kappa parameter.
               cl_fgs(spt_windows_lmin:spt_windows_lmax) = cl_fgs(spt_windows_lmin:spt_windows_lmax) - DataParams(6)*cl_derivative(spt_windows_lmin:spt_windows_lmax, 3)
            end if
            if (k==3) then
               !k=3 is TT spectrum.  Include BB Poisson foreground.
               cl_fgs(spt_windows_lmin:spt_windows_lmax) = 0.0
            end if
          
            !Now add model CMB.                                                                                 
            if (k==1) then !TE case                                                                                                                                                      
               cl_fgs(2:lmax) = cl_fgs(2:lmax) + Theory%cl(2:lmax, k+1) !TT is first, so skip it by adding 1 to k.                                   
            endif

            if (k==2) then !EE case                                                                                                                                                      
               cl_fgs(2:lmax) = cl_fgs(2:lmax) + Theory%cl(2:lmax, k+1) !TT is first, so skip it by adding 1 to k.                     
            endif

            if (k==3) then !BB case                                                                                                                                                      
               cl_fgs(2:lmax) = cl_fgs(2:lmax) + Theory%cl(2:lmax, k+1) !TT is first, so skip it by adding 1 to k.                     
            endif
          
            !Convert Cls to Dls
            dl_th(:) = cl_fgs(spt_windows_lmin:spt_windows_lmax)*cl_to_dl_conversion(spt_windows_lmin:spt_windows_lmax)
          
            !Now bin into bandpowers with the window functions.
            call dgemv('T',spt_windows_lmax-spt_windows_lmin+1,nbin,1.0d0,&
                 windows(:,1+nbin*(i-1):nbin+nbin*(i-1)),spt_windows_lmax-spt_windows_lmin+1,&
                 dl_th,1,0d0,tmpcb,1)

            !Finally, get \Delta Dl's between our model Dls and our input spectra
            !for covariance calculation.  Scale the data Dls by the right factors of Tcal and Pcal.
            
            !If i==1, spec(:,1) = SPTpol_TE.  Scale by Tcal and Pcal and add the TT leakage (epsilon*TT) to the theory bandpowers.
            if (i==1) then
               !First add the (scaled) data TT leakage to the theory expectation (same and subtracting from data).
               tmpcb(:) = tmpcb(:) + DataParams(5) * (DataParams(1)**2.0d0)*spec(:,3)
               
               !Now calculate delta between theory (tmpcb) and the scaled data.
               deltacb(1+(i-1)*nbin:nbin+(i-1)*nbin) = tmpcb(:) - (DataParams(1)**2.0d0) * DataParams(2) * spec(:,i)
            endif

            !If i==2, spec(:,2) = SPTpol_EE.  Scale by Tcal and Pcal and add the TT leakage (epsilon^2 * TT) to the theory bandpowers.
            if (i==2) then
               !First add the (scaled) data TT leakage to the theory expectation (same and subtracting from data).
               tmpcb(:) = tmpcb(:) + (DataParams(5)**2.0d0) * (DataParams(1)**2.0d0) * spec(:,3)

               !Now calculate delta between theory (tmpcb) and the scaled data.
               deltacb(1+(i-1)*nbin:nbin+(i-1)*nbin) = tmpcb(:) - (DataParams(1)**2.0d0) * (DataParams(2)**2.0d0) * spec(:,i)
            endif
         end do
      end do


      !We have to scale cov_data by the appropriate calibration parameters.
      do i=1,nall
         !If this is the TE portion, only hit with Tcal^2 Pcal.
         if (i <= nbin) then
            cal_vector(i) = DataParams(1)**2.0d0 * DataParams(2)
         endif

         !IF this is EE, hit with Tcal^2Pcal^2.
         if (i > nbin) then
            cal_vector(i) = DataParams(1)**2.0d0 * DataParams(2)**2.0d0
         endif
      end do

      !Now take the outer product of pcal_vector to get the correct
      !polarization calibration
      cov_scale = SPREAD(cal_vector, 2, nall) * SPREAD(cal_vector, 1, nall)      
      new_cov_data = cov_scale * cov_data

      !Now add the scaled cov_data to cov_sims.
      cov = cov_sims + new_cov_data

      !Cholesky decompose the total cov before calculating chi^2.
      call Matrix_CholeskyDouble(cov)

      !Now calculate the SPTpol likelihood.
      SPTpolLnLike = Matrix_GaussianLogLikeCholDouble(cov,deltacb) 
      !print *, 'SPTpolLnLike lnlike = ', SPTpolLnLike

      !Add Gaussian prior for temperature calibration.
      TcalLnLike = 0.5*((DataParams(1) - meanTcal)**2.0d0 /sigmaTcal**2.0d0 + log(sigmaTcal**2.0d0))

      !Add Gaussian prior for kappa.
      kappaLnLike = 0.5*((DataParams(6) - meankappa)**2.0d0 /sigmakappa**2.0d0 + log(sigmakappa**2.0d0))

      SPTpolLnLike = SPTpolLnLike + TcalLnLike + kappaLnLike

      if (feedback > 1)      print *, 'SPTpolLnLike lnlike = ', SPTpolLnLike
    end function SPTpolLnLike

    end module SPTpol_2012_TEEE
